\ **********************************************************************
\ * ADDMULTI.FTH fr F-TP 1.00  [ Turbo-FORTH-83 fr Transputer T800 ] *
\ *                                                                    *
\ *   Ergnzungen fr den Verkehr untereinander in einem Multisystem   *
\ *                                                                    *
\ *    Fred Behringer  Planegger Str. 24  81241 Mnchen    26.4.96     *
\ **********************************************************************

\ Dieses Programm wird von  MAKEMULT.BAT  beim Aufbau eines Multisystems
\ in jedes einzelne Subsystem hinzuINCLUDEt. Es gestattet, von Subsystem
\ zu Subsystem zu springen.  Auf einem Transputersystem ist das einfach:
\ Es braucht nur der Workspace-Pointer umgestellt zu werden.

DECIMAL
ONLY FORTH ALSO DEFINITIONS

\ Die Subsysteme  sollen alle mit dem Wort EMPTY abschlieen.  Also wird
\ FORTHKEY-MAKE  mit dem damit verbundenen Mechanismus zur Erzeugung von
\ EMPTY zunchst einmal herausgenommen und nachher wieder drangehngt.

DECIMAL
' FORTHKEY-MAKE >VIEW          \ Erstes Byte von FORTHKEY-MAKE
DUP FENCE !                    \ nach FENCE
(FORGET)                       \ und dann FORTHKEY-MAKE FORGETten

\ Jetzt erfolgt die Erweiterung zu einem Subsystem in einem Multisystem.
\ Die einzelnen Worte  werden von den Batch-Dateien  zur Erzeugung  des
\ Multisystems in Anspruch genommen.

HEX
 VARIABLE #SUBSYS
DOS TRANSMUL   ANZAHL!  BYE
40 0FE @-H DUP #SUBSYS !       \ Anzahl der Subsysteme
2VARIABLE SUBSYS   4* ALLOT    \ Tabelle fr die Sub-WPTR-Adressen
 VARIABLE OFS-DP               \ DP-(WPTR), zur Ermittlung der Sub-HEREs

: THERE ( n -- ad )  \ holt die absolute Adresse von HERE von Subsystem n
     DUP 1 #SUBSYS @ BETWEEN
     IF
       4* SUBSYS + @  OFS-DP @ + @       \ = Inhalt von DP im Subsystem n
     ELSE
       CR BOLD ." Es existiert kein solches Subsystem."
       CR      ." Zur Verfgung stehen 1 bis " #SUBSYS @ .
       CR ATTOFF  DROP
     THEN ;

\ Mit THERE sind Parameterbertragungen von Subsystem zu Subsystem Tr und
\ Tor geffnet.  Man beachte, dass man bei bertragungen  von Zahlenwerten
\ ( und auch sonst normalerweise ) gut daran tut, dem THERE ein EVEN nach-
\ zuschicken, um auf "Wortgrenze" ( 32-Bit-Grenze ) zu "gltten".

: SAVE-MULTI ( <filename[.ext]> )         \ Gesamtmultisystem ins Arbeits-
     SUBSYS 4+ @ #SUBSYS @ THERE SAVE ;   \ verzeichnis speichern

CODE (GOSUB) ( n -- )       \ n = (WPTR) des Subsystems
     SP LDL  0 LDNL         \ Adresse vom Stack holen
     SP LDL  4 ADC  SP STL  \ Stack anpassen
     GAJW                   \ Workspace-Pointer umstellen
     'QUIT LDL IP STL       \ cfa von QUIT des Zielsubsystems
     'QUIT LDL  W STL       \ nach IP und W .
     NESTP LDL  GCALL       \ Nach NEST springen
     END-CODE               \ NEXT nicht ntig

: GOSUB ( n -- )  \ springt zu Subsystem 0 < n < 256
     DUP 1 #SUBSYS @ BETWEEN
     IF
       4* SUBSYS + @ (GOSUB)
     ELSE
       CR BOLD ." Es existiert kein solches Subsystem."
       CR      ." Zur Verfgung stehen 1 bis " #SUBSYS @ .
       CR ATTOFF  DROP
     THEN ;

: SUBWPTR@ ( -- )
     #SUBSYS @ 0
     ?DO
       I 1+ 40 0FA !-H
       " TRANSMUL   SUBWPTR! BYE" SHELL
       40 0FE @-H SUBSYS 4+ I 4* +     CC!
       40 0FC @-H SUBSYS 4+ I 4* + 2 + CC!
     LOOP ;

SUBWPTR@        \ Die WPTR-Adressen der Subsysteme in die Tabelle SUBSYS legen.
FORGET SUBWPTR@ \ Und schon hat SUBWPTR@ seine Schuldigkeit getan.

DP WPTR@ -      \ Offset von DP zum Workspace-Pointer zwischenspeichern,
OFS-DP !        \ um an HERE in den einzelnen Subsystemen zu gelangen.

\ Und jetzt werden FORTHKEY-MAKE , (FORTHKEY) und EMPTY wieder eingebunden.

HEX

: FORTHKEY-MAKE ( n -- )                \ n = Lnge von FORTHKEY
     DEPTH 0<=
     ABORT" Lnge ?"                    \ Vergessen ?
     EVEN                               \ Aufgerundet. Jetzt durch 4 teilbar.
     DUP 20 <
     ABORT" Lnge negativ oder zu klein !"
     DUP SP0 @ PAD - >
     ABORT" Lnge zu gro !"
     " (FORTHKEY)" DROP 1- FIND         \ Schon da ?
     IF
       >VIEW DUP FENCE ! (FORGET)       \ Ja, dann zusammen mit EMPTY entfernen
     ELSE
       DROP                             \ Nein, dann Adresse vom Stack nehmen
     THEN
     " VARIABLE (FORTHKEY)" $EXECUTE
     " ' (FORTHKEY) IS FORTHKEY" $EXECUTE
     DUP HERE + FORTHKEY !              \ Tabellenende
     ALLOT                              \ Tabellenlnge
     FORTHKEY 8 + DUP OFF               \ Initiierung erstes "Record" = Lnge 0
     FORTHKEY 4+  !                     \ Zeiger auf erstes "Record"
     EMPTY$ $EXECUTE                    \ EMPTY erzeugen
                                        \ Aufruf von EMPTY reduziert
                                        \ das System auf KERNEL+TCORE .
                                        \ Das Wort EMPTY bleibt.
     " HERE ' EMPTY 2DUP 8 + ! 18 + !"  \ Diese Adresse,  das ist  die er-
     $EXECUTE                           \ ste Adresse nach dem Wort EMPTY,
                                        \ wird jetzt in das Wort EMPTY  so
                                        \ eingebracht, dass bei Aufruf von
                                        \ EMPTY  die  Variable  FENCE  auf
                                        \ diese Adresse  gesetzt  und  an-
                                        \ schlieend alles,  was auf EMPTY
                                        \ folgt, (FORGET)et wird.
                                        \ Siehe auch  KERNEL  am Ende  von
                                        \ TKERNEL.FTH.
     " EMPTY" $EXECUTE ;                \ Damit wird FENCE  auf den ersten
                                        \ Platz hinter EMPTY gesetzt.

800 FORTHKEY-MAKE                       \ Und schon wird das alles gemacht.


: INIT ( -- )
     #SUBSYS @ 1 >
     IF
       #SUBSYS @ 1+ 2 ?DO SUBSYS I 4* + @ HANDLE WPTR@ - + OFF LOOP
     THEN ;

\ INIT wird in MAKEMULT.BAT aufgerufen und nach Abarbeiten per EMPTY besei-
\ tigt.  Damit wird in allen neu (meta-)compilierten Subsystemen HANDLE OFF
\ gesetzt.  Das kann erst  zum Schluss gemacht werden,  da andernfalls  die
\ Batchdatei MAKEMULT.BAT in ....GOTO-SYSTEM-BOOT^ ...INCLUDE...BYE-FOR-DOS
\ keine Rcksprnge finden wrde..


ECHO OFF WARNING ON

HEX
