\ LSTRING.FTH

\ *******************************************************
\ *                                                     *
\ *  F-TP 1.00  [ Turbo-FORTH-83 fr Transputer T800 ]  *
\ *                                                     *
\ *      Utility - LSTRINGs als berlange STRINGs       *
\ *                                                     *
\ *  Fred Behringer, Planegger Str. 24,  81241 Mnchen  *
\ *                        21.1.1996                    *
\ *                                                     *
\ *******************************************************


\ STRING,  TYPE,  COUNT usw.  sind nur  auf eine maximale
\ Lnge von 255 d ausgelegt. Auerdem sind Strings norma-
\ lerweise  nicht  zur Aufnahme aller Zeichen vorgesehen.
\ Beispielsweise  macht das Anfhrungszeichen  Schwierig-
\ keiten.  Mit LSTRINGs  sollen "STRINGs"  bereitgestellt
\ werden,  in die "alles" hineingepackt werden kann. Bei-
\ spielsweise auch ganze Quelltextdateien,  die dann ber
\ ein ( noch zu schaffendes ) Wort EXECUTE-L$ nach Belie-
\ ben in das System hineincompiliert werden knnen ( run-
\ time-gesteuerte (Meta-)Compilation ).
\
\ Folgende Worte aus dem  Umfeld STRING  lassen sich auch
\ fr LSTRING anwenden:
\ TYPE >TYPE PARSE RIGHT$ LEFT$ MID$ ITEM INSERT


: LSTRING ( n -- <name-l$> beim Compilieren )
     CREATE DUP , 0 , DP +! HERE EVEN DP ! ;CODE
       W LDL  4 ADC       W STL  SP LDL  -8 ADC  SP STL
       W LDL  8 ADC      SP LDL   1 STNL
       W LDL  4 ADC  0 LDNL
                         SP LDL   0 STNL  NEXT   END-CODE

CODE LCOUNT ( ad -- ad+4 len )  \ = COUNT fr LSTRINGs, wie LENGTH
     SP LDL  0 LDNL  0 LDNL
     SP LDL  0 LDNL  4 ADC  SP LDL  0 STNL  1PUSH  END-CODE

: LOAD-L$  ( <name[.ext]> ad-l$ len -- ) \ Datei in LSTRING l$ laden
     DROP DUP 8 - @ OVER SWAP           \ ad ad lenmax
     LOAD                               \ ad len'
     SWAP 4- ! ;                        \ ad len'

: SAVE-L$  ( <name[.ext]> ad-l$ len -- ) \ LSTRING l$ in Datei speichern
     OVER + SAVE ;

: PLACE-L$ ( ad len to -- ) 3DUP 4+ SWAP MOVE ! DROP ;

: LEN-L$   ( ad len -- ad len lenmax ) OVER 8 - @ ;

: INSERT-L$ ( str strvar pos -- )
     0 MAX OVER MIN  -ROT LEN-L$ SWAP 4 PICK + OVER MIN
     2 PICK 4- !  2 PICK - >R + R> INSERT  ;

: APPEND-L$ ( ad1 len1 ad2 len2 -- )  DUP INSERT-L$ ;

: ERASE-L$  ( ad len pos l -- )
     >R 0 MAX OVER MIN  2DUP - R> MIN 0 MAX  2 PICK OVER - 4 PICK 4- !
     ROT OVER - 2 PICK -  >R -ROT + TUCK + SWAP  R> CMOVE ;

: INPUT$    ( strvar -- ) OVER >R LEN-L$ NIP EXPECT SPAN @ R> 4- ! ;






\ Es folgt ein Beispiel zur Anwendung von LSTRING und den
\ Zubringerworten SAVE-L$ LOAD-L$  usw.  Das Beispiel ist
\ vielleicht nicht besonders sinnvoll,  es zeigt aber mit
\ einfachen Mitteln  die  sich ergebenden  Mglichkeiten.
\ Um das Beispiel mit den brigen Worten dieser Datei per
\ INCLUDE LSTRING.FTH  einladbar zu machen,  versetze man
\ das gleich folgende  eof  hinter das folgende Beispiel.


: COPY ( <name1.ext> <name2.ext> -- )
  " 3000 LSTRING COPY1 COPY1 LOAD-L$ COPY1 SAVE-L$ FORGET COPY1"
  $EXECUTE ;


eof
****************************************************************
****************************************************************



: STRING ( n -- <namevar$> beim Compilieren )
     CREATE DUP C, 0 C, DP +! HERE EVEN DP ! ;USES DOSTRING ,

LABEL DOSTRING   \ 1 Byte max, 1 Byte count
       W LDL  4 ADC       W STL  SP LDL  -8 ADC  SP STL
       W LDL  2 ADC      SP LDL   1 STNL
       W LDL  1 ADC  LB  SP LDL   0 STNL  NEXT   END-CODE


: $! ( str strvar -- ) DROP 1- DUP 1- C@ ROT MIN SWAP PLACE ;

CODE COUNT ( ad -- ad+1 len )
     SP LDL  0 LDNL  LB  SP LDL  0 LDNL  1 ADC  SP LDL  0 STNL  1PUSH  END-CODE
CODE LENGTH ( ad -- ad+4 len )  \ = COUNT fr Strings mit Lnge in 4 Byte
     SP LDL  0 LDNL  0 LDNL
     SP LDL  0 LDNL  4 ADC  SP LDL  0 STNL  1PUSH  END-CODE

CR .( Stringbearbeitung )

: LEN$    ( ad len -- ad len lenmax)  OVER 2- C@ ;
: INPUT$ ( strvar ---)  OVER >R LEN$ NIP EXPECT SPAN @ R> 1- C! ;
: RIGHT$  ( ad1 len1 l -- ad2 len2 )  0 MAX OVER MIN >R + R@ - R> ;
: LEFT$   ( ad1 len1 l -- ad2 len2 )  0 MAX MIN ;
: MID$    ( ad1 len1 pos l -- ad2 len2 )  >R OVER SWAP - RIGHT$ R> LEFT$ ;
: ITEM  ( ad len n -- adn lenn )
     1 MAX >R TUCK R> 0
     DO   ROT  DROP BL SKIP OVER SWAP  BL SCAN  LOOP
     DROP OVER - ;
     \ Angenommen, der bei ad len in ITEM beginnende String enthlt
     \ mehrere "Worte", die durch  BL  getrennt sind.  Dann liefert
     \ ad len n ITEM  mit  adn lenn  die Adresse und Lnge des nten
     \ "Wortes" im String. Dabei wirken 0, -1 usw wie 1 und grere
     \ n als mglich wie das grtmgliche n.
: INSERT   ( ad1 len1 ad2 len2 -- )
     ROT OVER MIN >R R@ - OVER DUP R@ + ROT CMOVE> R> CMOVE ;
: INSERT$  ( str strvar pos -- )
     0 MAX OVER MIN  -ROT LEN$ SWAP 4 PICK + OVER MIN
     2 PICK 1- C!  2 PICK - >R + R> INSERT  ;
: APPEND$  ( ad1 len1 ad2 len2 -- )  DUP INSERT$ ;
: ERASE$   ( ad len pos l -- )
     >R 0 MAX OVER MIN  2DUP - R> MIN 0 MAX  2 PICK OVER - 4 PICK 1- C!
     ROT OVER - 2 PICK -  >R -ROT + TUCK + SWAP  R> CMOVE ;
VARIABLE FOUND
: SCAN-1ST ( a n c --- a n )  CAPS @ IF DROP ELSE SCAN THEN ;
: SEARCH   ( strad strlen bad blen -- n f )
     FOUND OFF  SWAP >R  2DUP U>=
     IF OVER - 1+ 2 PICK C@ R@ -ROT >R
        BEGIN R@ SCAN-1ST DUP
          IF >R 3DUP SWAP COMPARE 0=  IF FOUND ON R> DROP 0 >R THEN  R> THEN
          DUP
        WHILE 1 /STRING
        REPEAT R> 2DROP -ROT
     THEN 2DROP R> - FOUND @ ;
: SEARCH$ ( str strvar -- pos len )
     2 PICK >R SEARCH IF R> ELSE R> 2DROP 0 0 THEN ;
: DELETE$ ( str strvar -- ) 2DUP >R >R SEARCH$ R> R> 2SWAP ERASE$ ;

: LOAD ( <name[.ext]> ad len -- len' ) \ Datei nach ad ins TSystem laden
     FILENAME 0 (OPEN) ?DOS-ERR >R
     DUP 0>
     IF
       OVER SWAP                     \ ad ad' len  ad':=ad
       BEGIN                         \ ad ad' len
         2DUP                        \ ad ad' len ad' len
         8000 <
         IF                          \ ad ad' len ad'
           OVER                      \ ad ad' len  ad' len
         ELSE                        \ ad ad' len ad'
           SWAP 8000 - SWAP 8000     \ ad ad' len ad' 8000  len:=len-8000
         THEN
         DSEGMENT 1000 + 0 ROT R@
         GET-HOST                    \ 32K einladen und
         ?DOS-ERR TUCK               \ rberschaufeln
         DSEGMENT 1000 + -ROT 0 -ROT
         CMOVE-H>T                   \ ad ad' len len'
         ROT OVER + -ROT             \ ad ad'+len' len len'
         8000 =                      \ ad ad' len       ad':=ad'+len'
         OVER 0> AND                 \ ad ad' len fl
         IF                          \ ad ad' len
           FALSE                     \ weiter ...
         ELSE                        \ ad ad' len
           DROP SWAP -               \ len' (insgesamt)
           TRUE
         THEN                        \ LOAD ist das Gegenstck zu SAVE. An-
       UNTIL                         \ fangsstck der Lnge  len  der Datei
     ELSE                            \ name.ext nach ad legen. len' ist die
       2DROP 0                       \ tatschliche Lnge,  die weniger als
     THEN                            \ len  betragen kann,  wenn  die Datei
     R> (CLOSE) ;                    \ krzer als len ist.  Keine Beschrn-
                                     \ kung auf 64 K !!!  Abgesichert gegen
                                     \ Eingabe von len <= 0 .
